/*
 * File Name     : vdce_send.c
 *
 * Function Name : vdce_send
 *
 * Description   : send data to children tasks through socket
 *
 * Input(s)      : mytaskid - source task id
 *                 no_of_argc - 
 *                 arglist -
 *                 taskinfo - data routing information
 *                 no_of_data - no of data that are involving this application
 *
 * Output(s)     : sending data to specified tasks
 *
 * Programmed by : Ilkyeun Ra , Syracuse University
 *
*/

#include "vdce.h"
#include "dvm.h"
#include "dvm_msg.h"
#include "inet.h"
#include <sys/stat.h>
#include <fcntl.h>


extern	struct	task *_dvminfo;
extern	_total_data;
extern	FILE	*jfp;

extern 	int _cvmsockfd;
extern 	VDCE_MSG_HEADER _cvm_header;

vdce_send(port,buf,nsize) 
VDCE_PORT  port;
VDCE_BUF  *buf;
int nsize;
{

  int rc;
  int index, i;
  int sockdptr;

 char msg[1024]; 
 int         cvmsockfd;
 VDCE_MSG_HEADER   msg_header;
 char        *recv_msg, command[5], msg_data[2048];

/* Checking management channel message */

    cvmsockfd = _cvmsockfd;

    if (readn_avail(&cvmsockfd)) {
        recv_data_using_msg(cvmsockfd, &msg_header, &recv_msg);
#ifdef DEBUG
        fprintf(jfp," vdce_send :received %s\n ",recv_msg);
	fflush(jfp);
#endif
        decode_command(recv_msg, command, msg_data);

}

#ifdef DEBUG
        fprintf(jfp," in vdce_send :port=%d, matsize = %d, %d\n ",port, (int )*buf, (int *)buf);
	for (i = 0 ; i< nsize; i++) fprintf(jfp,"%c", buf[i]);
	fflush(jfp);
#endif

  /* 
   * Search for port information using port id
   */

  index = vdce_get_port(port,VDCESEND);

  if (index == -1) {
	fprintf(jfp,
        "vdce_send:%d port-id is not defined in your application\n",port);
	rc = 0;
	return(rc);
  }

  /*
   * Check a type of destination 
   */
  if (_dvminfo[index].v_type == 0) 
        /* get data from an external file */
	rc = send_data_to_file(&_dvminfo[index],buf);
  else 
        /* get data from a task via socket */
        rc = send_msg_to_task(&_dvminfo[index],buf, nsize);
/*        rc = send_data_to_task(&_dvminfo[index],buf);
*/


  if (rc==0) {
     sprintf(msg,"VDCE_SEND: can't send data %s to %s\n",
		     _dvminfo[index].s_name,_dvminfo[index].d_name);
     err_ret(msg);
  }

#ifdef DEBUG
        fprintf(jfp, "vdce_send done\n");
        fflush(jfp);
#endif
  sockdptr = _dvminfo[index].s_port;
  /* close(sockdptr); */
  return(rc);

} /* end vdce_send */

send_msg_to_task(taskinfo,buf, nsize)
struct task *taskinfo;
char        *buf;
int 	    nsize;
{
     VDCE_MSG_HEADER hmsg;
     char        *message;
     char        *msgptr;
     int         rc, i;
     int         nwrite,nleft,sizeofdata,header_size,msgsize;
     int         sockfd;
 
     /*
      * read data from memory buf and write them to socket
      */
 
      sockfd     = taskinfo->s_socket;   
			/* assign socket descriptor */
      if (sockfd == -1) {
         err_ret("vdce_send: can't find socket descriptor");
         rc =0;
         return(rc);
      }
 
/*
      sizeofdata = strlen(buf);
*/
      sizeofdata = nsize;
 
#ifdef SENDDEBUG
      fprintf(jfp,"TOBETX:n=%d:%s\n",sizeofdata,buf);
      fflush(jfp);
#endif
      /*
       * write message header to inform size of data to be send to rec
eivcer */
      strcpy(hmsg.H_name, HEADER_MESSAGE);
      strcpy(hmsg.message_type,DVM2DVM);
      bzero(hmsg.destination_addr, DUMMYSIZE);
      create_header_address(&hmsg.destination_addr[9], 
		taskinfo->user_id, taskinfo->appl_id, 
		taskinfo->d_tkid);
      bzero(hmsg.source_addr, DUMMYSIZE);
      create_header_address(&hmsg.source_addr[9],
		taskinfo->user_id, taskinfo->appl_id, 
		taskinfo->s_tkid);
 
/*
      sprintf(hmsg.length, "%d",strlen(buf)+HEADERSIZE);
 buf is going to structure, neec check size of and strlen
*/
      sprintf(hmsg.length, "%d",nsize + HEADERSIZE);
 
#ifdef SENDDEBUG
/* just for matrix test
{
 typedef struct Matrix_struct {
  int   matrix_size;
  float matrix[16][16];
} matrix;
 
  matrix *show_matrix;
  int i, j;

 show_matrix = (struct Matrix_struct *)buf;
 
  printf("matrix size : %d\n", show_matrix->matrix_size);
  for ( i = 0 ; i < 16 ; i++ ) {
    for ( j = 0 ; j < 16 ; j++ ) {
      printf("%f  ", show_matrix->matrix[i][j]);
    }
    printf("\n");
  }

}
*/
        fprintf(jfp, "send buf size+header= %s\n", hmsg.length);
/*
        for(i = 0; i < nsize ; i++)
		fprintf(jfp, "%c", buf[i]);
*/
        fflush(jfp);
#endif
      send_data_using_msg(sockfd, hmsg, buf, nsize);

 
} 

/*
 * Function Name : send_data_to_task
 *
 * Description   : Sending data via socket
 *
 * Input(s)      : 
 *                 taskinfo - information of task execution enviroment
 *                 buf      - memory buffer of task result
 *
 *
*/
send_data_to_task(taskinfo,buf)
struct task *taskinfo;
char	    *buf;
{
     VDCE_HEADER hmsg;
     char	 *message;
     char	 *msgptr;
     int         rc;
     int         nwrite,nleft,sizeofdata,header_size,msgsize;
     int         sockfd;

     /*
      * read data from memory buf and write them to socket
      */

      sockfd     = taskinfo->s_socket;   /* assign socket descriptor */
      if (sockfd == -1) {
	 err_ret("vdce_send: can't find socket descriptor");
	 rc =0;
	 return(rc);
      }

      sizeofdata = strlen(buf); /* get size of data to be sent */

#ifdef SENDDEBUG 
      fprintf(jfp,"TOBETX:n=%d:%s\n",sizeofdata,buf);
#endif
      /* 
       * write message header to inform size of data to be send to receivcer
      */
      hmsg.proc    = taskinfo->s_tkid; 	   /* source task id */
      hmsg.msgtype = taskinfo->v_type;     /* data type      */
      hmsg.size    = sizeofdata;           /* size of data   */
      header_size  = sizeof(VDCE_HEADER);  /* size of header */

      
      /*   
       * create memory buffer for sending data 
      */

      msgsize     = header_size + hmsg.size;
      message      = (char *) malloc(msgsize);
      bzero(message,msgsize);
      msgptr	   = (char *) malloc(1*sizeof(int));

      if (message == NULL) {
	err_ret(jfp,"vdce_send: no memory!");
	rc = 0;
        return(rc);
      }

     msgptr = message;

      bcopy(&hmsg,message,header_size); /* copy header of data */
      bcopy(buf,message+header_size,sizeofdata); /* copy data */

#ifdef SENDDEBUG
     {
       int kk;
       fprintf(jfp,"VDCE_SEND(%d): %d size of data will be sent to Task %d:\n",
	taskinfo->s_tkid,msgsize,taskinfo->d_tkid);
/*
	for(kk=0;kk<msgsize;kk++)
        fprintf(jfp,"%c",msgptr[kk]);
       fprintf(jfp,"\n");
*/
     }
#endif


      /*
       * Loop until whole data is written  
      */
      nleft  = msgsize;

      while ( nleft > 0) {

	    nwrite = write(sockfd,message,nleft);

            if (nwrite <=0 ) {
                close(sockfd);
                err_ret("vdce_send: Error in writing data to a socekt");
	        rc = 0;
		return(rc);
             }
             nleft  -= nwrite;
	     message += nwrite;
      }   

      rc = 1;
      sleep(1);    /* for Sync */
#ifdef DEBUG
fprintf(jfp,"VDCE_SEND:(%d) task sent %d byte(s) data to task %d via socket %d\n",
        taskinfo->s_tkid,msgsize,taskinfo->d_tkid,sockfd);
#endif
      return(rc);
} /* end send_data_to_task */

/*
 * Function Name : send_data_to_file
 *
 * Description   : Sending data to a file
 *
 * Input(s)      : 
 *                 taskinfo - information of task execution enviroment
 *                 buf      - memory buffer of task result
 *
 *
*/
send_data_to_file(taskinfo,buf)
struct task *taskinfo;
char        *buf;
{
     int rc;
     int nwrite,nleft,sizeofdata;
     int filefd;
     char *eof;

     /*
      * read data from file buf and write them to memory buffer
      */

      filefd = open(&taskinfo->v_name[0],
               O_WRONLY|O_CREAT|O_TRUNC,0777);   /* get file descriptor */
      if (filefd == -1) {
	 fprintf(jfp,"vdce_send: can't creat %s file of result\n",
                         taskinfo->v_name);
	 rc =0;
	 return(rc);
      }

      sizeofdata = strlen(buf); /* get size of data to be sent */

      nwrite = 0;
      nleft  = sizeofdata;

      while ( nleft > 0) {

            nwrite = write(filefd,buf,nleft);

            if (nwrite <=0 ) {
                fprintf(jfp,"vdce_send: Error in writing data to a file %s\n",
                        taskinfo->v_name);
                close(filefd);
                rc = 0;
                return(rc);
             }
             nleft  -= nwrite;
             buf    += nwrite;
      }

/*
      eof = (char *) malloc(2);
      bzero(eof,2);
      eof = "\n";
      write(filefd,eof,strlen(eof));
*/
      close(filefd);
      rc = 1;
#ifdef DEBUG
fprintf(jfp,"VDCE_SEND:(%d) task stored %d byte(s) data at %s\n",
        taskinfo->s_tkid,sizeofdata,taskinfo->v_name);
      fflush(jfp);
#endif
      return(rc);

} /* end of send_data_to_file */

send_data_using_msg(sockfd, hmsg, buf, nsize)
int sockfd;
VDCE_MSG_HEADER hmsg;
char	    *buf;
int nsize;
{
     char	 *message;
     char	 *msgptr;
     int	 i;
     int         rc;
     int         nwrite,nleft,sizeofdata,header_size,msgsize;

     /*
      * read data from memory buf and write them to socket
      */

      if (sockfd == -1) {
	 err_ret("vdce_send: can't find socket descriptor");
	 rc =0;
	 return(rc);
      }

/*
      sizeofdata = strlen(buf); 
*/
      sizeofdata = nsize; 

#ifdef SENDDEBUG 
      fprintf(jfp,"TOBETX:n=%d:%s Header=%d\n",sizeofdata,buf, HEADERSIZE);
     fflush(jfp);
#endif
      /* 
       * write message header to inform size of data to be send to receivcer
      */
      
      /*   
       * create memory buffer for sending data 
      */

      msgsize     = HEADERSIZE + sizeofdata;
      message      = (char *) malloc(msgsize+1);
      bzero(message,msgsize+1);
      msgptr	   = (char *) malloc(1*sizeof(int));

      if (message == NULL) {
	err_ret("vdce_send: no memory!");
	rc = 0;
        return(rc);
      }

     msgptr = message;

      bcopy(&hmsg,message,HEADERSIZE); /* copy header of data */
      bcopy(buf,message+HEADERSIZE,sizeofdata); /* copy data */

#ifdef SENDDEBUG
     {
       int kk;
       fprintf(jfp,"send_msg(): %d size message %s\n",
	msgsize,message);
/*
	for(kk=0;kk<msgsize;kk++)
       	   if ((msgptr[kk]== '\0') || (msgptr[kk] ==' '))
                fprintf(jfp, " ");
         	else fprintf(jfp,"%c",msgptr[kk]);

       fprintf(jfp,"\n");
*/
	fflush(jfp);
     }
#endif


      /* for(i=0;i<5;i++) { */
	      /*
	       * Loop until whole data is written  
	      */
	      nleft  = msgsize;
	      message = msgptr;
	      while ( nleft > 0) {

		    nwrite = write(sockfd,message,nleft);

		    if (nwrite <=0 ) {
			close(sockfd);
			err_ret("send_msg: Error in writing data to a socket");
			rc = 0;
			return(rc);
		     }
              nleft  -= nwrite;
	      message += nwrite;
	     }   

	      rc = 1;
	      sleep(3);    /* for Sync */
      /* } */
      return(rc);
} /* end send_data_using_msg */
